/* _/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
   _/          = USB driver of TTP. =        _/
   _/ gunze.c : Main source-code             _/
   _/ First : 2002/04/16  Touch : 2004/10/19 _/
   _/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
   _/   Copyright (c) 2000-2002 GUNZE Ltd    _/
   _/          All rights reserved           _/
   _/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
[Compile]
gcc -D__KERNEL__ -I/usr/src/linux/include -Wall -O2 -DMODULE -DMODVERSIONS -include /usr/src/linux/include/linux/modversions.h -c -o gunze.o gunze.c
******************************************************************************/
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
#include<linux/kernel.h>
#include<linux/slab.h>
#include<linux/module.h>
#include<linux/init.h>
#include<linux/input.h>
#include<linux/usb.h>
#include<asm/uaccess.h>
#define USB_VENDOR_ID_GUNZE  0x0637  /*                       */
#define USB_PRODUCT_ID_GUNZE 0x0001  /* GUNZE USB Information */
#define USB_VERSION_ID_GUNZE 0x0100  /*                       */
#define DRIVER_VERSION "Version 4.0.0"     /*                    */
#define DRIVER_AUTHOR  "GUNZE Ltd."        /* Driver Information */
#define DRIVER_DESC    "for Input System"  /*                    */
MODULE_AUTHOR( DRIVER_AUTHOR );
MODULE_DESCRIPTION( DRIVER_DESC ": " DRIVER_VERSION);
/*****************************************************************************/
struct GUNZE_TTP{
  int                iOpenNum;
  unsigned char      ucpData[4];
  struct input_dev   sIDev;
  struct urb         sIrq;      /* USB Request Block */
  struct usb_device *spUDev;
};
static int giDevmajor=60;
struct GUNZE_TTP *gspGunzeTtp;
/*****************************************************************************/
static void gunze_ttp_irq(struct urb *spUrb){
  unsigned char    *ucpData=gspGunzeTtp->ucpData;
  struct input_dev *spIDev=&gspGunzeTtp->sIDev;
  if(spUrb->status) return;
  input_report_key(spIDev, BTN_LEFT, ucpData[0]&0x20);
  input_report_abs(spIDev, ABS_X, ucpData[2]|(((__u32)ucpData[0]&0x1F) << 7));
  input_report_abs(spIDev, ABS_Y, ucpData[3]|(((__u32)ucpData[1]&0x1F) << 7));
  input_event(spIDev, EV_RST, 0, 0);
}
/*****************************************************************************/
static void gunze_ttp_cmd(struct urb *spUrb){
  struct usb_api_data *spAwd;
  spAwd=(struct usb_api_data *)spUrb->context;
  spAwd->done=1;
  wmb();
  wake_up(&spAwd->wqh);
}
/*****************************************************************************/
static int gunze_ttp_open(struct input_dev *spIDev){
  if(gspGunzeTtp->iOpenNum++) return 0;
  gspGunzeTtp->sIrq.dev=gspGunzeTtp->spUDev;
  if(usb_submit_urb(&gspGunzeTtp->sIrq)) return -EIO;
  return 0;
}
/*****************************************************************************/
static void gunze_ttp_close(struct input_dev *spIDev){
  if(!--gspGunzeTtp->iOpenNum) usb_unlink_urb(&gspGunzeTtp->sIrq);
}
/*****************************************************************************/
static void *gunze_ttp_probe(struct usb_device *spUDev, unsigned int uiIfNum,
                                         const struct usb_device_id *spUDevId){
  int iPipe, iMaxp;
  struct GUNZE_TTP                *spGunzeTtp;
  struct usb_interface            *spIf;
  struct usb_interface_descriptor *spIfDesc;
  struct usb_endpoint_descriptor  *spEpDesc;
  if((spUDev->descriptor.idVendor != USB_VENDOR_ID_GUNZE) ||
     (spUDev->descriptor.idProduct != USB_PRODUCT_ID_GUNZE) ||
     (spUDev->descriptor.bcdDevice < USB_PRODUCT_ID_GUNZE) ||
     (spUDev->descriptor.bNumConfigurations != 1)) return NULL;
  spIf=&spUDev->actconfig->interface[uiIfNum];
  spIfDesc=&spIf->altsetting[spIf->act_altsetting];
  spEpDesc=spIfDesc->endpoint;
  if((!(spEpDesc->bEndpointAddress&0x80))||
     ((spEpDesc->bmAttributes&3) != 3)) return NULL;
  iPipe=usb_rcvintpipe(spUDev, spEpDesc->bEndpointAddress);
  iMaxp=usb_maxpacket(spUDev, iPipe, usb_pipeout(iPipe));
  usb_set_idle(spUDev, spIfDesc->bInterfaceNumber, 0, 0);
  if(!(spGunzeTtp=kmalloc(sizeof(struct GUNZE_TTP), GFP_KERNEL))) return NULL;
  memset(spGunzeTtp, 0, sizeof(struct GUNZE_TTP));
  spGunzeTtp->sIDev.evbit[0]=BIT(EV_KEY)|BIT(EV_ABS)|BIT(EV_RST);
  spGunzeTtp->sIDev.keybit[LONG(BTN_LEFT)]|=BIT(BTN_LEFT);
  spGunzeTtp->sIDev.absbit[0]|=BIT(ABS_X)|BIT(ABS_Y);
  spGunzeTtp->sIDev.absmax[ABS_X]=spGunzeTtp->sIDev.absmax[ABS_Y]=4095;
  spGunzeTtp->sIDev.private=   spGunzeTtp;
  spGunzeTtp->sIDev.open=      gunze_ttp_open;
  spGunzeTtp->sIDev.close=     gunze_ttp_close;
  spGunzeTtp->sIDev.idbus=     BUS_USB;
  spGunzeTtp->sIDev.idvendor=  spUDev->descriptor.idVendor;
  spGunzeTtp->sIDev.idproduct= spUDev->descriptor.idProduct;
  spGunzeTtp->sIDev.idversion= spUDev->descriptor.bcdDevice;
  spGunzeTtp->spUDev=spUDev;
  gspGunzeTtp=spGunzeTtp;
  FILL_INT_URB(&spGunzeTtp->sIrq, spUDev, iPipe, spGunzeTtp->ucpData,
        iMaxp > 8 ? 8 : iMaxp, gunze_ttp_irq, spGunzeTtp, spEpDesc->bInterval);
  input_register_device(&spGunzeTtp->sIDev);
  printk(KERN_INFO "input%d: gunze_ttp on usb%d:%d.%d\n",
       spGunzeTtp->sIDev.number, spUDev->bus->busnum, spUDev->devnum, uiIfNum);
  return spGunzeTtp;
}
/*****************************************************************************/
static void gunze_ttp_disconnect(struct usb_device *spUDev, void *vpArg){
  usb_unlink_urb(&gspGunzeTtp->sIrq);
  input_unregister_device(&gspGunzeTtp->sIDev);
  kfree(gspGunzeTtp);
}
/*****************************************************************************/
static struct usb_device_id gsGunzeTtpIdTable[]={
  {bInterfaceClass:0xFF, bInterfaceSubClass:0xFF, bInterfaceProtocol:0xFF},
  {}
};
/*****************************************************************************/
MODULE_DEVICE_TABLE(usb, gsGunzeTtpIdTable);
static struct usb_driver gunze_ttp_driver={
  name:       "gunze_ttp",
  probe:      gunze_ttp_probe,
  disconnect: gunze_ttp_disconnect,
  id_table:   gsGunzeTtpIdTable,
};
/*****************************************************************************/
static int gunze_ttp_ioctl(struct inode *spInode, struct file *spFile,
                                      unsigned int uiCmd, unsigned long ulArg){
  /* Req10h -> ST:0 SI:1 SU:2 SD:3, Req11h -> DR:0 DW:1 EA:2 CA:3 MD:4 MN:5 */ 
  /* uiCmd:    0    1    2    3               4    5    6    7    8    9 */
  /* Req12h -> MR, uiCmd=10 val=address_of_controller index=0 */
  int iPipe;
  __u8  ucRequestType=0xC1, ucRequest=0x10;
  __u16 usValue=0x00, usIndex=0, usSize=2;
  char  cpTmpBuffer[2];
  cpTmpBuffer[0]=0;
  if(uiCmd >= 0 && uiCmd < 4) usValue=uiCmd;
  else if(uiCmd >= 4 && uiCmd < 10){
    ucRequest=0x11;
    usValue=uiCmd-4;
  }else if(uiCmd == 10/*MR*/){
    unsigned char ucAddr[2];
    ucRequest=0x12;
    copy_from_user((void *)ucAddr, (void *)ulArg, 2);
    usValue=(unsigned char)ucAddr[0];
  }
  iPipe=usb_rcvctrlpipe(gspGunzeTtp->spUDev, 0);
  {
    struct urb *spUrb;
    DECLARE_WAITQUEUE(wait, current);
    struct usb_api_data spAwd;
    int iStatus, iTimeout=200;
    struct usb_ctrlrequest *spDr
                          =kmalloc(sizeof(struct usb_ctrlrequest), GFP_KERNEL);
    if(!spDr) return -ENOMEM;
    spDr->bRequestType=ucRequestType;
    spDr->bRequest=ucRequest;
    spDr->wValue=cpu_to_le16p(&usValue);
    spDr->wIndex=cpu_to_le16p(&usIndex);
    spDr->wLength=cpu_to_le16p(&usSize);
    if(!(spUrb=usb_alloc_urb(0))) return -ENOMEM;
    FILL_CONTROL_URB(spUrb, gspGunzeTtp->spUDev, iPipe, (unsigned char *)spDr,
                                     cpTmpBuffer, 2, gunze_ttp_cmd, NULL);
    init_waitqueue_head(&spAwd.wqh);  
    spAwd.done=0;
    set_current_state(TASK_UNINTERRUPTIBLE);
    add_wait_queue(&spAwd.wqh, &wait);
    spUrb->context=&spAwd;
    iStatus=usb_submit_urb(spUrb);
    if(iStatus){
      usb_free_urb(spUrb);
      set_current_state(TASK_RUNNING);
      remove_wait_queue(&spAwd.wqh, &wait);
      printk("Error: Can't submit urb.\n");
      return -EIO;
    }
    while(iTimeout && !spAwd.done){
      iTimeout=schedule_timeout(iTimeout);
      set_current_state(TASK_UNINTERRUPTIBLE);
      rmb();
    }
    set_current_state(TASK_RUNNING);
    remove_wait_queue(&spAwd.wqh, &wait);
    if(!iTimeout && !spAwd.done){
      if(spUrb->status != -EINPROGRESS){
        printk("usb: raced timeout, "
                            "pipe 0x%x status %d time left %d\n",
                            spUrb->pipe, spUrb->status, iTimeout);
        iStatus=spUrb->status;
      }else{
         printk("usb_control/bulk_msg: timeout ----\n");
         usb_unlink_urb(spUrb);
         iStatus=-ETIMEDOUT;
      }
    }else iStatus=spUrb->status;
    /*  printk("ReqTyp=%xh Req=%xh Val=%xh Idx=%xh Len=%xh\n",
       ((struct usb_ctrlrequest *)spUrb->setup_packet)->bRequestType,
       ((struct usb_ctrlrequest *)spUrb->setup_packet)->bRequest,
       ((struct usb_ctrlrequest *)spUrb->setup_packet)->wValue,
       ((struct usb_ctrlrequest *)spUrb->setup_packet)->wIndex,
       ((struct usb_ctrlrequest *)spUrb->setup_packet)->wLength); */
    usb_free_urb(spUrb);
    copy_to_user((void *)ulArg, cpTmpBuffer, 2);
    kfree(spDr);
  }
  return 0;
}
/*****************************************************************************/
static int gunze_ttp_open2(struct inode *spInode, struct file *spFile){
  MOD_INC_USE_COUNT; return 0;
}
/*****************************************************************************/
static int gunze_ttp_close2(struct inode *spInode, struct file *spFile){
  MOD_DEC_USE_COUNT; return 0;
}
/*****************************************************************************/
static struct file_operations gunze_ttp_fops={
  open:    gunze_ttp_open2,
  release: gunze_ttp_close2,
  ioctl:   gunze_ttp_ioctl,
};
/*****************************************************************************/
static int __init gunze_ttp_init(void){
  usb_register(&gunze_ttp_driver);
  if(!giDevmajor){
    if((giDevmajor=register_chrdev(0, "gunze", &gunze_ttp_fops)) < 0){
      printk("No available device node.\n"); return -EBUSY;
    }
  }else if(register_chrdev(giDevmajor, "gunze", &gunze_ttp_fops) < 0){
    printk("Device Registration Error.\n"); return -EBUSY;
  }
  info(DRIVER_DESC ": " DRIVER_VERSION);
  return 0;
}
/*****************************************************************************/
static void __exit gunze_ttp_exit(void){
  usb_deregister(&gunze_ttp_driver);
  if(unregister_chrdev(giDevmajor, "gunze")) printk("Unregistration Error.\n");
}
/*****************************************************************************/
module_init(gunze_ttp_init);
module_exit(gunze_ttp_exit);
