/* _/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
   _/          = USB driver of TTP. =        _/
   _/ gunze.c : for linux-2.6.x              _/
   _/ First : 2003/12/24  Touch : 2004/10/19 _/
   _/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
   _/   Copyright (c) 2000-2004 GUNZE Ltd    _/
   _/          All rights reserved           _/
   _/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
[Compile] Use our package.
************************************************************************* */
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
#include<linux/kernel.h>
#include<linux/slab.h>
#include<linux/module.h>
#include<linux/init.h>
#include<linux/input.h>
#include<linux/usb.h>
#include<asm/uaccess.h>
#define USB_VENDOR_ID_GUNZE  0x0637  /*                       */
#define USB_PRODUCT_ID_GUNZE 0x0001  /* GUNZE USB Information */
#define USB_VERSION_ID_GUNZE 0x0100  /*                       */
#define DRIVER_VERSION "Version 4.0.0"     /*                    */
#define DRIVER_AUTHOR  "GUNZE Ltd."        /* Driver Information */
#define DRIVER_DESC    "for Input System"  /*                    */
#define DRIVER_LICENSE "GPL"
MODULE_AUTHOR( DRIVER_AUTHOR );
MODULE_DESCRIPTION( DRIVER_DESC ": " DRIVER_VERSION);
MODULE_LICENSE(DRIVER_LICENSE);
/*****************************************************************************/
struct GUNZE_TTP{
  int                iOpenNum;
  unsigned char      *ucpData;
  struct input_dev   sIDev;
  struct urb         *spIrq;      /* USB Request Block */
  struct usb_device  *spUDev;
  char               cPhys[64];
  dma_addr_t         sDataDma;
};
static int giDevmajor=60;
struct GUNZE_TTP *gspGunzeTtp;
struct usb_api_data{
  wait_queue_head_t wqh;
  int done;
};
/*****************************************************************************/
static void gunze_ttp_irq(struct urb *spUrb, struct pt_regs *spRegs){
  unsigned char    *ucpData=gspGunzeTtp->ucpData;
  struct input_dev *spIDev=&gspGunzeTtp->sIDev;
  if(spUrb->status) return;
  input_regs(spIDev, spRegs);
  input_report_key(spIDev, BTN_LEFT, ucpData[0]&0x20);
  input_report_abs(spIDev, ABS_X, ucpData[2]|(((__u32)ucpData[0]&0x1F) << 7));
  input_report_abs(spIDev, ABS_Y, ucpData[3]|(((__u32)ucpData[1]&0x1F) << 7));
  input_sync(spIDev);
  usb_submit_urb(spUrb, SLAB_ATOMIC);
}
/*****************************************************************************/
static void gunze_ttp_cmd(struct urb *spUrb){
  struct usb_api_data *spAwd;
  spAwd=(struct usb_api_data *)spUrb->context;
  spAwd->done=1;
  wmb();
  wake_up(&spAwd->wqh);
}
/*****************************************************************************/
static int gunze_ttp_open(struct input_dev *spIDev){
  if(gspGunzeTtp->iOpenNum++) return 0;
  gspGunzeTtp->spIrq->dev=gspGunzeTtp->spUDev;
  if(usb_submit_urb(gspGunzeTtp->spIrq, GFP_KERNEL)) return -EIO;
  return 0;
}
/*****************************************************************************/
static void gunze_ttp_close(struct input_dev *spIDev){
  if(!--gspGunzeTtp->iOpenNum) usb_unlink_urb(gspGunzeTtp->spIrq);
}
/*****************************************************************************/
static int gunze_ttp_probe(struct usb_interface *spIfNum,
                             const struct usb_device_id *spUDevId){
  int    iPipe, iMaxp;
  char   cpPath[64];
  struct GUNZE_TTP                *spGunzeTtp;
  struct usb_host_interface       *spIf;
  struct usb_endpoint_descriptor  *spEpDesc;
  struct usb_device               *spUDev=interface_to_usbdev(spIfNum);
  if((spUDev->descriptor.idVendor != USB_VENDOR_ID_GUNZE) ||
     (spUDev->descriptor.idProduct != USB_PRODUCT_ID_GUNZE) ||
     (spUDev->descriptor.bcdDevice < USB_PRODUCT_ID_GUNZE) ||
     (spUDev->descriptor.bNumConfigurations != 1)) return -ENODEV;
  spIf=&spIfNum->altsetting[spIfNum->act_altsetting];
  if(spIf->desc.bNumEndpoints != 1) return -ENODEV;
  spEpDesc=&spIf->endpoint[0].desc;
  if((!(spEpDesc->bEndpointAddress&0x80))||
     ((spEpDesc->bmAttributes&3) != 3)) return -ENODEV;
  iPipe=usb_rcvintpipe(spUDev, spEpDesc->bEndpointAddress);
  iMaxp=usb_maxpacket(spUDev, iPipe, usb_pipeout(iPipe));
  if(!(spGunzeTtp=kmalloc(sizeof(struct GUNZE_TTP), GFP_KERNEL)))
    return -ENODEV;
  memset(spGunzeTtp, 0, sizeof(struct GUNZE_TTP));
  spGunzeTtp->ucpData=usb_buffer_alloc(spUDev, 4, SLAB_ATOMIC,
                                       &spGunzeTtp->sDataDma);
  if(!spGunzeTtp->ucpData){ kfree(spGunzeTtp); return -ENODEV; }
  spGunzeTtp->spIrq=usb_alloc_urb(0, GFP_KERNEL);
  if(!spGunzeTtp->spIrq){
    usb_buffer_free(spUDev, 4, spGunzeTtp->ucpData, spGunzeTtp->sDataDma);
    kfree(spGunzeTtp);
    return -ENODEV;
  }
  spGunzeTtp->spUDev=spUDev;
  spGunzeTtp->sIDev.evbit[0]=BIT(EV_KEY)|BIT(EV_ABS);
  spGunzeTtp->sIDev.keybit[LONG(BTN_LEFT)]|=BIT(BTN_LEFT);
  spGunzeTtp->sIDev.absbit[0]|=BIT(ABS_X)|BIT(ABS_Y);
  spGunzeTtp->sIDev.absmax[ABS_X]=spGunzeTtp->sIDev.absmax[ABS_Y]=4095;
  spGunzeTtp->sIDev.private=   spGunzeTtp;
  spGunzeTtp->sIDev.open=      gunze_ttp_open;
  spGunzeTtp->sIDev.close=     gunze_ttp_close;
  usb_make_path(spUDev, cpPath, 64);
  sprintf(spGunzeTtp->cPhys, "%s/input0", cpPath);
  spGunzeTtp->sIDev.id.bustype=     BUS_USB;
  spGunzeTtp->sIDev.id.vendor=  spUDev->descriptor.idVendor;
  spGunzeTtp->sIDev.id.product= spUDev->descriptor.idProduct;
  spGunzeTtp->sIDev.id.version= spUDev->descriptor.bcdDevice;
  gspGunzeTtp=spGunzeTtp;
  usb_fill_int_urb(spGunzeTtp->spIrq, spUDev, iPipe, spGunzeTtp->ucpData,
        iMaxp > 8 ? 8 : iMaxp, gunze_ttp_irq, spGunzeTtp, spEpDesc->bInterval);
  spGunzeTtp->spIrq->transfer_dma=spGunzeTtp->sDataDma;
  spGunzeTtp->spIrq->transfer_flags|=URB_NO_TRANSFER_DMA_MAP;
  input_register_device(&spGunzeTtp->sIDev);
  printk(KERN_INFO "input: gunze_ttp on %s\n", cpPath);
  usb_set_intfdata(spIfNum, spGunzeTtp);
  return 0;
}
/*****************************************************************************/
static void gunze_ttp_disconnect(struct usb_interface *spIfNum){
  usb_set_intfdata(spIfNum, NULL);
  usb_unlink_urb(gspGunzeTtp->spIrq);
  input_unregister_device(&gspGunzeTtp->sIDev);
  usb_free_urb(gspGunzeTtp->spIrq);
  usb_buffer_free(interface_to_usbdev(spIfNum), 4, gspGunzeTtp->ucpData,
                  gspGunzeTtp->sDataDma);
  kfree(gspGunzeTtp);
}
/*****************************************************************************/
static struct usb_device_id gsGunzeTtpIdTable[]={
  { USB_INTERFACE_INFO(0xFF, 0xFF, 0xFF) },
  {}
};
/*****************************************************************************/
MODULE_DEVICE_TABLE(usb, gsGunzeTtpIdTable);
static struct usb_driver gunze_ttp_driver={
  .owner      =THIS_MODULE,
  .name       ="gunze_ttp",
  .probe      =gunze_ttp_probe,
  .disconnect =gunze_ttp_disconnect,
  .id_table   =gsGunzeTtpIdTable,
};
/*****************************************************************************/
static int gunze_ttp_ioctl(struct inode *spInode, struct file *spFile,
                                      unsigned int uiCmd, unsigned long ulArg){
  /* Req10h -> ST:0 SI:1 SU:2 SD:3, Req11 -> DR:0 DW:1 EA:2 CA:3 MD:4 MN:5 */
  /* uiCmd:    0    1    2    3              4    5    6    7    8    9 */ 
  /* Req12h -> MR, uiCmd=10 val=address_of_controller index=0 */
  int iPipe;
  __u8  ucRequestType=0xC1, ucRequest=0x10;
  __u16 usValue=0x00, usIndex=0, usSize=2;
  char  cpTmpBuffer[2];
  cpTmpBuffer[0]=0;
  if(uiCmd >= 0 && uiCmd < 4) usValue=uiCmd;
  else if(uiCmd >= 4 && uiCmd < 10){
    ucRequest=0x11;
    usValue=uiCmd-4;
  }else if(uiCmd == 10/*MR*/){
    unsigned char ucAddr[2];
    ucRequest=0x12;
    copy_from_user((void *)ucAddr, (void *)ulArg, 2);
    usValue=(unsigned char)ucAddr[0];
  }
  iPipe=usb_rcvctrlpipe(gspGunzeTtp->spUDev, 0);
  {
    struct urb *spUrb;
    DECLARE_WAITQUEUE(wait, current);
    struct usb_api_data spAwd;
    int iStatus, iTimeout=200;
    struct usb_ctrlrequest *spDr;
    dma_addr_t sDrDma;
    if(!(spDr=usb_buffer_alloc(gspGunzeTtp->spUDev,
	sizeof(struct usb_ctrlrequest), SLAB_ATOMIC, &sDrDma))) return -ENOMEM;
    if(!spDr) return -ENOMEM;
    spDr->bRequestType=ucRequestType;
    spDr->bRequest=ucRequest;
    spDr->wValue=cpu_to_le16p(&usValue);
    spDr->wIndex=cpu_to_le16p(&usIndex);
    spDr->wLength=cpu_to_le16p(&usSize);
    if(!(spUrb=usb_alloc_urb(0,GFP_KERNEL))) return -ENOMEM;
    usb_fill_control_urb(spUrb, gspGunzeTtp->spUDev, iPipe,
                         (unsigned char *)spDr, cpTmpBuffer, 2,
                         (usb_complete_t)gunze_ttp_cmd, NULL);
    init_waitqueue_head(&spAwd.wqh);  
    spAwd.done=0;
    set_current_state(TASK_UNINTERRUPTIBLE);
    add_wait_queue(&spAwd.wqh, &wait);
    spUrb->context=&spAwd;
    iStatus=usb_submit_urb(spUrb, SLAB_ATOMIC);
    if(iStatus){
      usb_free_urb(spUrb);
      set_current_state(TASK_RUNNING);
      remove_wait_queue(&spAwd.wqh, &wait);
      printk("Error: Can't submit urb.\n");
      return -EIO;
    }
    while(iTimeout && !spAwd.done){
      iTimeout=schedule_timeout(iTimeout);
      set_current_state(TASK_UNINTERRUPTIBLE);
      rmb();
    }
    set_current_state(TASK_RUNNING);
    remove_wait_queue(&spAwd.wqh, &wait);
    if(!iTimeout && !spAwd.done){
      if(spUrb->status != -EINPROGRESS){
        printk("usb: raced timeout, "
                            "pipe 0x%x status %d time left %d\n",
                            spUrb->pipe, spUrb->status, iTimeout);
        iStatus=spUrb->status;
      }else{
         printk("usb_control/bulk_msg: timeout ----\n");
         usb_unlink_urb(spUrb);
         iStatus=-ETIMEDOUT;
      }
    }else iStatus=spUrb->status;
    /*  printk("ReqTyp=%xh Req=%xh Val=%xh Idx=%xh Len=%xh\n",
       ((struct usb_ctrlrequest *)spUrb->setup_packet)->bRequestType,
       ((struct usb_ctrlrequest *)spUrb->setup_packet)->bRequest,
       ((struct usb_ctrlrequest *)spUrb->setup_packet)->wValue,
       ((struct usb_ctrlrequest *)spUrb->setup_packet)->wIndex,
       ((struct usb_ctrlrequest *)spUrb->setup_packet)->wLength); */
    usb_free_urb(spUrb);
    copy_to_user((void *)ulArg, cpTmpBuffer, 2);
    if(spDr) usb_buffer_free(gspGunzeTtp->spUDev,
                             sizeof(struct usb_ctrlrequest), spDr, sDrDma);
  }
  return 0;
}
/*****************************************************************************/
static int gunze_ttp_open2(struct inode *spInode, struct file *spFile){
  return 0;
}
/*****************************************************************************/
static int gunze_ttp_close2(struct inode *spInode, struct file *spFile){
  return 0;
}
/*****************************************************************************/
static struct file_operations gunze_ttp_fops={
  .open =    gunze_ttp_open2,
  .release = gunze_ttp_close2,
  .ioctl =   gunze_ttp_ioctl,
};
/*****************************************************************************/
static int __init gunze_ttp_init(void){
  usb_register(&gunze_ttp_driver);
  if(!giDevmajor){
    if((giDevmajor=register_chrdev(0, "gunze", &gunze_ttp_fops)) < 0){
      printk("No available device node.\n"); return -EBUSY;
    }
  }else if(register_chrdev(giDevmajor, "gunze", &gunze_ttp_fops) < 0){
    printk("Device Registration Error.\n"); return -EBUSY;
  }
  info(DRIVER_DESC ": " DRIVER_VERSION);
  return 0;
}
/*****************************************************************************/
static void __exit gunze_ttp_exit(void){
  usb_deregister(&gunze_ttp_driver);
  if(unregister_chrdev(giDevmajor, "gunze")) printk("Unregistration Error.\n");
}
/*****************************************************************************/
module_init(gunze_ttp_init);
module_exit(gunze_ttp_exit);
